"""Device capabilities."""

from __future__ import annotations

import logging
from enum import IntEnum
from typing import Any

_LOGGER = logging.getLogger(__name__)


class DeviceCapability(IntEnum):
    """Available device capabilities."""

    EDGE_CUT = 1
    ONE_TIME_SCHEDULE = 2
    PARTY_MODE = 4
    TORQUE = 8
    OFF_LIMITS = 16
    CUTTING_HEIGHT = 32
    ACS = 64


CAPABILITY_TO_TEXT = {
    DeviceCapability.EDGE_CUT: "Edge Cut",
    DeviceCapability.ONE_TIME_SCHEDULE: "One-Time-Schedule",
    DeviceCapability.PARTY_MODE: "Party Mode",
    DeviceCapability.TORQUE: "Motor Torque",
    DeviceCapability.OFF_LIMITS: "Off Limits",
    DeviceCapability.CUTTING_HEIGHT: "Cutting Height",
    DeviceCapability.ACS: "ACS",
}


class Capability:
    """Class for handling device capabilities."""

    def __init__(self, device_data: Any | None = None) -> int:
        """Initialize the capability list."""
        # super().__init__()
        self.__int__: int = 0
        self.ready: bool = False
        if isinstance(device_data, type(None)):
            return

        cfg = (
            device_data["cfg"]
            if "cfg" in device_data
            else device_data["last_status"]["payload"]["cfg"]
        )
        dat = (
            device_data["dat"]
            if "dat" in device_data
            else device_data["last_status"]["payload"]["dat"]
        )

        try:
            if "sc" in cfg:
                if "ots" in cfg["sc"] or "once" in cfg["sc"]:
                    self.add(DeviceCapability.ONE_TIME_SCHEDULE)
                    self.add(DeviceCapability.EDGE_CUT)

                if "distm" in cfg["sc"] or "enabled" in cfg["sc"]:
                    self.add(DeviceCapability.PARTY_MODE)

        except TypeError:
            pass

        try:
            if "modules" in dat:
                # Offlimits module
                if "DF" in dat["modules"]:
                    self.add(DeviceCapability.OFF_LIMITS)

                # Set cutting height
                if "EA" in dat["modules"]:
                    self.add(DeviceCapability.CUTTING_HEIGHT)

                # ACS module
                if "US" in dat["modules"]:
                    self.add(DeviceCapability.ACS)
        except TypeError:
            pass

        try:
            if "tq" in cfg:
                self.add(DeviceCapability.TORQUE)
        except TypeError:
            pass

    def add(self, capability: DeviceCapability) -> None:
        """Add capability to the list."""
        if capability & self.__int__ == 0:
            self.__int__ = self.__int__ | capability

    def check(self, capability: DeviceCapability) -> bool:
        """Check if device has capability."""
        if capability & self.__int__ == 0:
            return False
        else:
            return True



"""pyWorxCloud definition."""

# pylint: disable=undefined-loop-variable
# pylint: disable=line-too-long
# pylint: disable=too-many-lines
from __future__ import annotations

import json
import logging
import sys
import threading
from datetime import datetime, timedelta
from random import randint
from typing import Any
from zoneinfo import ZoneInfo

from .api import LandroidCloudAPI
from .clouds import CloudType
from .events import EventHandler, LandroidEvent
from .exceptions import (
    AuthorizationError,
    MowerNotFoundError,
    NoACSModuleError,
    NoConnectionError,
    NoCuttingHeightError,
    NoOfflimitsError,
    NoOneTimeScheduleError,
    NoPartymodeError,
    OfflineError,
    TooManyRequestsError,
    ZoneNoProbability,
    ZoneNotDefined,
)
from .helpers import convert_to_time, get_logger
from .utils import MQTT, DeviceCapability, DeviceHandler
from .utils.mqtt import Command
from .utils.requests import HEADERS, POST

if sys.version_info < (3, 9, 0):
    sys.exit("The pyWorxcloud module requires Python 3.9.0 or later")

_LOGGER = logging.getLogger(__name__)

API_REFRESH_TIME_MIN = 5
API_REFRESH_TIME_MAX = 10


class WorxCloud(dict):
    """
    Worx by Landroid Cloud connector.

    Used for handling API connection to Worx, Kress and Landxcape devices which are cloud connected.

    This uses a reverse engineered API protocol, so no guarantee that this will keep working.
    There are no public available API documentation available.
    """

    # __device: str | None = None

    def __init__(
        self,
        username: str,
        password: str,
        cloud: (
            CloudType.WORX | CloudType.KRESS | CloudType.LANDXCAPE | str
        ) = CloudType.WORX,
        verify_ssl: bool = True,
        tz: str | None = None,  # pylint: disable=invalid-name
    ) -> None:
        """
        Initialize :class:WorxCloud class and set default attribute values.

        1. option for connecting and printing the current states from the API, using :code:`with`

        .. testcode::
        from pyworxcloud import WorxCloud
        from pprint import pprint

        with WorxCloud("your@email","password","worx", 0, False) as cloud:
            pprint(vars(cloud))

        2. option for connecting and printing the current states from the API, using :code:`connect` and :code:`disconnect`

        .. testcode::
        from pyworxcloud import WorxCloud
        from pprint import pprint

        cloud = WorxCloud("your@email", "password", "worx")

        # Initialize connection
        auth = cloud.authenticate()

        if not auth:
            # If invalid credentials are used, or something happend during
            # authorize, then exit
            exit(0)

        # Connect to device with index 0 (devices are enumerated 0, 1, 2 ...)
        # and do not verify SSL (False)
        cloud.connect(0, False)

        # Read latest states received from the device
        cloud.update()

        # Print all vars and attributes of the cloud object
        pprint(vars(cloud))

        # Disconnect from the API
        cloud.disconnect()

        For further information, see the Wiki for documentation: https://github.com/MTrab/pyworxcloud/wiki

        Args:
            username (str): Email used for logging into the app for your device.
            password (str): Password for your account.
            cloud (CloudType.WORX | CloudType.KRESS | CloudType.LANDXCAPE | str, optional): The CloudType matching your device. Defaults to CloudType.WORX.
            index (int, optional): Device number if more than one is connected to your account (starting from 0 representing the first added device). Defaults to 0.
            verify_ssl (bool, optional): Should this module verify the API endpoint SSL certificate? Defaults to True.

        Raise:
            TypeError: Error raised if invalid CloudType was specified.
        """
        _LOGGER.debug("Initializing connector...")
        super().__init__()

        self._worx_mqtt_client_id = None

        if not isinstance(
            cloud,
            (
                type(CloudType.WORX),
                type(CloudType.LANDXCAPE),
                type(CloudType.KRESS),
            ),
        ):
            try:
                _LOGGER.debug("Try getting correct CloudType from %s", cloud.upper())
                cloud = getattr(CloudType, cloud.upper())
                _LOGGER.debug("Found cloud type %s", cloud)
            except AttributeError:
                raise TypeError(
                    "Wrong type specified, valid types are: worx, landxcape or kress"
                ) from None

        _LOGGER.debug("Initializing the API connector ...")
        self._api = LandroidCloudAPI(username, password, cloud, tz, self._token_updated)
        self._username = username
        self._cloud = cloud
        self._auth_result = False
        _LOGGER.debug("Getting logger ...")
        self._log = get_logger("pyworxcloud")
        self._raw = None
        self._tz = tz

        self._save_zones = None
        self._verify_ssl = verify_ssl
        _LOGGER.debug("Initializing EventHandler ...")
        self._events = EventHandler()

        self._endpoint = None
        self._user_id = None
        self._mowers = None

        self._decoding: bool = False

        # Dict holding refresh timers
        self._timers = {}

        # Dict of devices, identified by name
        self.devices: DeviceHandler = {}

        self.mqtt = None

    def __enter__(self) -> Any:
        """Default actions using with statement."""
        self.authenticate()

        self.connect()

        return self

    def __exit__(self, exc_type, exc_value, traceback) -> Any:
        """Called on end of with statement."""
        self.disconnect()

    def authenticate(self) -> bool:
        """Authenticate against the API."""
        self._log.debug("Authenticating %s", self._username)

        try:
            self._api.get_token()
        except TooManyRequestsError:
            raise TooManyRequestsError from None

        auth = self._api.authenticate()
        if auth is False:
            self._auth_result = False
            self._log.debug("Authentication for %s failed!", self._username)
            raise AuthorizationError("Unauthorized")

        self._auth_result = True
        self._log.debug("Authentication for %s successful", self._username)

        return True

    def update_attribute(self, device: str, attr: str, key: str, value: Any) -> None:
        """Used as callback to update value."""
        chattr = self.devices[device]
        if not isinstance(attr, type(None)):
            for level in attr.split(";;"):
                if hasattr(chattr, level):
                    chattr = getattr(chattr, level)
                else:
                    chattr = chattr[level]

        if hasattr(chattr, key):
            setattr(chattr, key, value)
        elif isinstance(chattr, dict):
            chattr.update({key: value})

    def set_callback(self, event: LandroidEvent, func: Any) -> None:
        """Set callback which is called when data is received.

        Args:
            event: LandroidEvent for this callback
            func: Function to be called.
        """
        self._events.set_handler(event, func)

    def disconnect(self) -> None:
        """Close API connections."""
        # pylint: disable=bare-except
        logger = self._log.getChild("Disconnect")

        # Cancel force refresh timer on disconnect
        try:
            for _, tmr in self._timers.items():
                tmr.cancel()
        except:
            logger.debug("Could not cancel timers - skipping.")

        # Disconnect MQTT connection
        try:
            self.mqtt.disconnect()
        except:
            logger.debug("Could not disconnect MQTT - skipping.")

    def connect(
        self,
    ) -> bool:
        """
        Connect to the cloud service endpoint

        Returns:
            bool: True if connection was successful, otherwise False.
        """
        self._log.debug("Fetching basic API data")
        self._fetch()
        self._log.debug("Done fetching basic API data")

        if len(self._mowers) == 0:
            self._log.debug("no mowers connected to account")
            return False

        self._endpoint = self._mowers[0]["mqtt_endpoint"]
        self._user_id = self._mowers[0]["user_id"]

        self._log.debug("Setting up MQTT handler")
        # setup MQTT handler
        self.mqtt = MQTT(
            self._api,
            self._cloud.BRAND_PREFIX,
            self._endpoint,
            self._user_id,
            self._log,
            self._on_update,
        )

        self.mqtt.connect()

        for mower in self._mowers:
            self.mqtt.subscribe(mower["mqtt_topics"]["command_out"], True)

        # Convert time strings to objects.
        for name, device in self.devices.items():
            convert_to_time(
                name, device, device.time_zone, callback=self.update_attribute
            )

        self._log.debug("Connection tasks all done")

        return True

    def _token_updated(self) -> None:
        """Called when token is updated."""
        self.mqtt.update_token()

    @property
    def auth_result(self) -> bool:
        """Return current authentication result."""
        return self._auth_result

    def _on_update(self, payload):  # , topic, payload, dup, qos, retain, **kwargs):
        """Triggered when a MQTT message was received."""
        logger = self._log.getChild("MQTT_data_in")
        try:
            data = json.loads(payload)
            logger.debug("MQTT data received")

            # "Malformed" message, we are missing a serial number and
            # MAC address to identify the mower.
            if (
                not "sn" in data["cfg"] and not "uuid" in data["dat"]
            ) and not "mac" in data["dat"]:
                logger.debug("Malformed message received")
                return

            found_match = False

            for mower in self._mowers:
                if "sn" in data["cfg"]:
                    if mower["serial_number"] == data["cfg"]["sn"]:
                        found_match = True
                        break
                elif "uuid" in data["dat"]:
                    if mower["uuid"] == data["dat"]["uuid"]:
                        found_match = True
                        break
                elif "mac" in data["dat"]:
                    if mower["mac_address"] == data["dat"]["mac"]:
                        found_match = True
                        break

            if not found_match:
                logger.debug("Could not match incoming data with a known mower!")
                return
            else:
                logger.debug("Matched to '%s'", mower["name"])

            device: DeviceHandler = self.devices[mower["name"]]

            if not device.online:
                logger.debug("Device is marked offline - refreshing")
                self._fetch()
                device: DeviceHandler = self.devices[mower["name"]]

            if "raw_data" in mower and mower["raw_data"] == data:
                self._log.debug("Data was already present and not changed.")
                return  # Dataset was not changed, no update needed

            mower["raw_data"] = data
            device: DeviceHandler = self.devices[mower["name"]]
            device.raw_data = data

            self._events.call(
                LandroidEvent.DATA_RECEIVED, name=mower["name"], device=device
            )
        except json.decoder.JSONDecodeError:
            logger.debug("Malformed MQTT message received")

    def _on_api_update(self, data):  # , topic, payload, dup, qos, retain, **kwargs):
        """Triggered when API has been updated."""
        logger = self._log.getChild("API_update")
        try:
            self._events.call(LandroidEvent.API, api_data=data)
        except json.decoder.JSONDecodeError:
            logger.debug("Malformed MQTT message received")

    def _fetch(self, forced: bool = False) -> None:
        """Fetch base API information."""
        self._mowers = self._api.get_mowers()
        # self.devices = {}
        for mower in self._mowers:
            try:
                device = DeviceHandler(self._api, mower, self._tz, False)
                if not isinstance(mower["last_status"], type(None)):
                    device.raw_data = mower["last_status"]["payload"]

                _LOGGER.debug("Mower '%s' data: %s", mower["name"], mower)
                self.devices.update({mower["name"]: device})

                if isinstance(mower["mac_address"], type(None)):
                    mower["mac_address"] = (
                        device.raw_data["dat"]["mac"]
                        if "mac" in device.raw_data["dat"]
                        else "__UUID__"
                    )

                if forced:
                    self._events.call(
                        LandroidEvent.API, name=mower["name"], device=device
                    )
            except TypeError:
                pass

        self._schedule_api_refresh()

    def _schedule_api_refresh(self) -> None:
        """Schedule the API refresh."""
        logger = self._log.getChild("API_Refresh_Scheduler")

        try:
            self._timers["api"].cancel()
        except KeyError:
            pass

        refresh_secs = (randint(API_REFRESH_TIME_MIN, API_REFRESH_TIME_MAX)) * 60
        timezone = (
            ZoneInfo(self._tz)
            if not isinstance(self._tz, type(None))
            else ZoneInfo("UTC")
        )
        now = datetime.now().astimezone(timezone)
        next_api_refresh = now + timedelta(seconds=refresh_secs)
        logger.debug(
            "Scheduling an API refresh at %s",
            next_api_refresh,
        )

        force_api_refresh = threading.Timer(refresh_secs, self._fetch, args=[True])
        force_api_refresh.start()
        self._timers.update({"api": force_api_refresh})

    def get_mower(self, serial_number: str, device: bool = False) -> dict:
        """Get a specific mower object.

        Args:
            serial_number (str): Serial number of the device
        """

        if device:
            for mower in self.devices.items():
                if mower[1].serial_number == serial_number:
                    return mower[1]
        else:
            for mower in self._mowers:
                if mower["serial_number"] == serial_number:
                    return mower

        raise MowerNotFoundError(
            f"Mower with serialnumber {serial_number} was not found."
        )

    def update(self, serial_number: str) -> None:
        """Request a state refresh."""
        mower = self.get_mower(serial_number)
        _LOGGER.debug("Trying to refresh '%s'", serial_number)

        try:
            self.mqtt.ping(
                serial_number if mower["protocol"] == 0 else mower["uuid"],
                mower["mqtt_topics"]["command_in"],
                mower["protocol"],
            )
        except NoConnectionError:
            raise NoConnectionError from None

    def start(self, serial_number: str) -> None:
        """Start mowing task

        Args:
            serial_number (str): Serial number of the device

        Raises:
            OfflineError: Raised if the device is offline.
        """
        mower = self.get_mower(serial_number)
        if mower["online"]:
            _LOGGER.debug("Sending start command to '%s'", serial_number)
            self.mqtt.command(
                serial_number if mower["protocol"] == 0 else mower["uuid"],
                mower["mqtt_topics"]["command_in"],
                Command.START,
                mower["protocol"],
            )
        else:
            raise OfflineError("The device is currently offline, no action was sent.")

    def home(self, serial_number: str) -> None:
        """Stop the current task and go home.
        If the knifes was turned on when this is called,
        it will return home with knifes still turned on.

        Args:
            serial_number (str): Serial number of the device

        Raises:
            OfflineError: Raised if the device is offline.
        """
        mower = self.get_mower(serial_number)

        if mower["online"]:
            self.mqtt.command(
                serial_number if mower["protocol"] == 0 else mower["uuid"],
                mower["mqtt_topics"]["command_in"],
                Command.HOME,
                mower["protocol"],
            )
        else:
            raise OfflineError("The device is currently offline, no action was sent.")

    def safehome(self, serial_number: str) -> None:
        """Stop and go home with the blades off

        Args:
            serial_number (str): Serial number of the device

        Raises:
            OfflineError: Raised if the device is offline.
        """
        mower = self.get_mower(serial_number)
        if mower["online"]:
            self.mqtt.command(
                serial_number if mower["protocol"] == 0 else mower["uuid"],
                mower["mqtt_topics"]["command_in"],
                Command.SAFEHOME,
                mower["protocol"],
            )
        else:
            raise OfflineError("The device is currently offline, no action was sent.")

    def pause(self, serial_number: str) -> None:
        """Pause the mowing task

        Args:
            serial_number (str): Serial number of the device

        Raises:
            OfflineError: Raised if the device is offline.
        """
        mower = self.get_mower(serial_number)
        if mower["online"]:
            self.mqtt.command(
                serial_number if mower["protocol"] == 0 else mower["uuid"],
                mower["mqtt_topics"]["command_in"],
                Command.PAUSE,
                mower["protocol"],
            )
        else:
            raise OfflineError("The device is currently offline, no action was sent.")

    def raindelay(self, serial_number: str, rain_delay: str) -> None:
        """Set new rain delay.

        Args:
            serial_number (str): Serial number of the device
            rain_delay (str): Rain delay in minutes.

        Raises:
            OfflineError: Raised if the device is offline.
        """
        mower = self.get_mower(serial_number)
        if mower["online"]:
            if not isinstance(rain_delay, int):
                rain_delay = int(rain_delay)
            self.mqtt.publish(
                serial_number if mower["protocol"] == 0 else mower["uuid"],
                mower["mqtt_topics"]["command_in"],
                {"rd": rain_delay},
            )
        else:
            raise OfflineError("The device is currently offline, no action was sent.")

    def set_lock(self, serial_number: str, state: bool) -> None:
        """Set the device locked state.

        Args:
            serial_number (str): Serial number of the device
            state (bool): True will lock the device, False will unlock the device.

        Raises:
            OfflineError: Raised if the device is offline.
        """
        mower = self.get_mower(serial_number)
        if mower["online"]:
            self.mqtt.command(
                serial_number if mower["protocol"] == 0 else mower["uuid"],
                mower["mqtt_topics"]["command_in"],
                Command.LOCK if state else Command.UNLOCK,
                mower["protocol"],
            )
        else:
            raise OfflineError("The device is currently offline, no action was sent.")

    def set_partymode(self, serial_number: str, state: bool) -> None:
        """Turn on or off the partymode.

        Args:
            serial_number (str): Serial number of the device
            state (bool): True is enabling partymode, False is disabling partymode.

        Raises:
            NoPartymodeError: Raised if the device does not support partymode.
            OfflineError: Raised if the device is offline.
        """
        mower = self.get_mower(serial_number)

        if mower["online"]:
            device = DeviceHandler(self._api, mower, self._tz)
            if device.capabilities.check(DeviceCapability.PARTY_MODE):
                if mower["protocol"] == 0:
                    self.mqtt.publish(
                        serial_number if mower["protocol"] == 0 else mower["uuid"],
                        mower["mqtt_topics"]["command_in"],
                        (
                            {"sc": {"m": 2, "distm": 0}}
                            if state
                            else {"sc": {"m": 1, "distm": 0}}
                        ),
                        mower["protocol"],
                    )
                else:
                    self.mqtt.publish(
                        serial_number if mower["protocol"] == 0 else mower["uuid"],
                        mower["mqtt_topics"]["command_in"],
                        {"sc": {"enabled": 0}} if state else {"sc": {"enabled": 1}},
                        mower["protocol"],
                    )
            elif not device.capabilities.check(DeviceCapability.PARTY_MODE):
                raise NoPartymodeError("This device does not support Partymode")
        elif not mower["online"]:
            raise OfflineError("The device is currently offline, no action was sent.")

    def set_offlimits(self, serial_number: str, state: bool) -> None:
        """Turn on or off the off limits module.

        Args:
            serial_number (str): Serial number of the device
            state (bool): True is enabling off limits module, False is disabling off limits module.

        Raises:
            NoOfflimitsError: Raised if the device does not support off limits.
            OfflineError: Raised if the device is offline.
        """
        mower = self.get_mower(serial_number)

        if mower["online"]:
            _LOGGER.debug("Setting offlimits")
            device = DeviceHandler(self._api, mower, self._tz)
            if device.capabilities.check(DeviceCapability.OFF_LIMITS):
                self.mqtt.publish(
                    serial_number if device.protocol == 0 else device.uuid,
                    mower["mqtt_topics"]["command_in"],
                    (
                        {
                            "modules": {
                                "DF": {
                                    "cut": 1,
                                    "fh": 1 if device.offlimit_shortcut else 0,
                                }
                            }
                        }
                        if state
                        else {
                            "modules": {
                                "DF": {
                                    "cut": 0,
                                    "fh": 1 if device.offlimit_shortcut else 0,
                                }
                            }
                        }
                    ),
                    device.protocol,
                )
            elif not device.capabilities.check(DeviceCapability.OFF_LIMITS):
                raise NoOfflimitsError("This device does not support Off Limits")
        elif not mower["online"]:
            raise OfflineError("The device is currently offline, no action was sent.")

    def set_offlimits_shortcut(self, serial_number: str, state: bool) -> None:
        """Turn on or off the off limits shortcut function.

        Args:
            serial_number (str): Serial number of the device
            state (bool): True is enabling shortcut, False is disabling shortcut.

        Raises:
            NoOfflimitsError: Raised if the device does not support off limits.
            OfflineError: Raised if the device is offline.
        """
        mower = self.get_mower(serial_number)

        if mower["online"]:
            _LOGGER.debug("Setting offlimits")
            device = DeviceHandler(self._api, mower, self._tz)
            if device.capabilities.check(DeviceCapability.OFF_LIMITS):
                self.mqtt.publish(
                    serial_number if device.protocol == 0 else device.uuid,
                    mower["mqtt_topics"]["command_in"],
                    (
                        {
                            "modules": {
                                "DF": {
                                    "cut": 1 if device.offlimit else 0,
                                    "fh": 1,
                                }
                            }
                        }
                        if state
                        else {
                            "modules": {
                                "DF": {
                                    "cut": 1 if device.offlimit else 0,
                                    "fh": 0,
                                }
                            }
                        }
                    ),
                    device.protocol,
                )
            elif not device.capabilities.check(DeviceCapability.OFF_LIMITS):
                raise NoOfflimitsError("This device does not support Off Limits")
        elif not mower["online"]:
            raise OfflineError("The device is currently offline, no action was sent.")

    def setzone(self, serial_number: str, zone: str | int) -> None:
        """Set zone to be mowed when next mowing task is started.

        Args:
            serial_number (str): Serial number of the device
            zone (str | int): Zone to mow, valid possibilities are a number from 1 to 4.

        Raises:
            OfflineError: Raised if the device is offline.
        """
        mower = self.get_mower(serial_number)
        if mower["online"]:
            device = DeviceHandler(self._api, mower, self._tz)
            if not isinstance(zone, int):
                zone = int(zone)

            if (
                zone >= len(device.zone["starting_point"])
                or device.zone["starting_point"][zone] == 0
            ):
                raise ZoneNotDefined(
                    f"Cannot request zone {zone} as it is not defined."
                )

            if not zone in device.zone["indicies"]:
                raise ZoneNoProbability(
                    f"Cannot request zone {zone} as it has no probability set."
                )

            current_zones = device.zone["indicies"]
            requested_zone_index = current_zones.index(zone)
            next_zone_index = device.zone["index"]

            no_indices = len(current_zones)
            offset = (requested_zone_index - next_zone_index) % no_indices
            new_zones = []
            for i in range(0, no_indices):
                new_zones.append(current_zones[(offset + i) % no_indices])

            device = DeviceHandler(self._api, mower, self._tz)
            self.mqtt.publish(
                serial_number if mower["protocol"] == 0 else mower["uuid"],
                mower["mqtt_topics"]["command_in"],
                {"mzv": new_zones},
                mower["protocol"],
            )
        else:
            raise OfflineError("The device is currently offline, no action was sent.")

    def zonetraining(self, serial_number: str) -> None:
        """Start the zone training task.

        Args:
            serial_number (str): Serial number of the device

        Raises:
            OfflineError: Raised if the device is offline.
        """
        mower = self.get_mower(serial_number)
        if mower["online"]:
            _LOGGER.debug("Sending ZONETRAINING command to %s", mower["name"])
            self.mqtt.command(
                serial_number if mower["protocol"] == 0 else mower["uuid"],
                mower["mqtt_topics"]["command_in"],
                Command.ZONETRAINING,
                mower["protocol"],
            )
        else:
            raise OfflineError("The device is currently offline, no action was sent.")

    def restart(self, serial_number: str):
        """Reboot the device baseboard.

        Args:
            serial_number (str): Serial number of the device

        Raises:
            OfflineError: Raised if the device is offline.
        """
        mower = self.get_mower(serial_number)
        if mower["online"]:
            _LOGGER.debug("Sending RESTART command to %s", mower["name"])
            self.mqtt.command(
                serial_number if mower["protocol"] == 0 else mower["uuid"],
                mower["mqtt_topics"]["command_in"],
                Command.RESTART,
                mower["protocol"],
            )
        else:
            raise OfflineError("The device is currently offline, no action was sent.")

    def toggle_schedule(self, serial_number: str, enable: bool) -> None:
        """Turn on or off the schedule.

        Args:
            serial_number (str): Serial number of the device
            enable (bool): True is enabling the schedule, Fasle is disabling the schedule.

        Raises:
            OfflineError: Raised if the device is offline.
        """
        mower = self.get_mower(serial_number)
        if mower["online"]:
            self.mqtt.publish(
                serial_number if mower["protocol"] == 0 else mower["uuid"],
                mower["mqtt_topics"]["command_in"],
                {"sc": {"m": 1}} if enable else {"sc": {"m": 0}},
                mower["protocol"],
            )
        else:
            raise OfflineError("The device is currently offline, no action was sent.")

    def ots(self, serial_number: str, boundary: bool, runtime: str) -> None:
        """Start a One-Time-Schedule task

        Args:
            serial_number (str): Serial number of the device
            boundary (bool): If True the device will start the task cutting the edge.
            runtime (str | int): Minutes to run the task before returning to dock.

        Raises:
            NoOneTimeScheduleError: OTS is not supported by the device.
            OfflineError: Raised when the device is offline.
        """
        mower = self.get_mower(serial_number)
        if mower["online"]:
            device = DeviceHandler(self._api, mower, self._tz)
            if device.capabilities.check(DeviceCapability.ONE_TIME_SCHEDULE):
                if not isinstance(runtime, int):
                    runtime = int(runtime)

                device = DeviceHandler(self._api, mower, self._tz)
                self.mqtt.publish(
                    serial_number if mower["protocol"] == 0 else mower["uuid"],
                    mower["mqtt_topics"]["command_in"],
                    {"sc": {"ots": {"bc": int(boundary), "wtm": runtime}}},
                    mower["protocol"],
                )
            elif not device.capabilities.check(DeviceCapability.ONE_TIME_SCHEDULE):
                raise NoOneTimeScheduleError(
                    "This device does not support Edgecut-on-demand"
                )
        else:
            raise OfflineError("The device is currently offline, no action was sent.")

    def send(self, serial_number: str, data: str) -> None:
        """Send raw JSON data to the device.

        Args:
            serial_number (str): Serial number of the device
            data (str): Data to be sent, formatted as a valid JSON object.

        Raises:
            OfflineError: Raised if the device isn't online.
        """
        mower = self.get_mower(serial_number)
        if mower["online"]:
            _LOGGER.debug("Sending %s to %s", data, mower["name"])
            self.mqtt.publish(
                serial_number if mower["protocol"] == 0 else mower["uuid"],
                mower["mqtt_topics"]["command_in"],
                json.loads(data),
                mower["protocol"],
            )
        else:
            raise OfflineError("The device is currently offline, no action was sent.")

    def reset_charge_cycle_counter(self, serial_number: str) -> None:
        """Resets charge cycle counter.

        Args:
            serial_number (str): Serial number of the device
            data (str): Data to be sent, formatted as a valid JSON object.

        Raises:
            OfflineError: Raised if the device isn't online.
        """
        mower = self.get_mower(serial_number)
        if mower["online"]:
            _LOGGER.debug("Resetting charge cycle counter for %s", mower["name"])
            self._api.check_token()
            POST(
                f"https://{self._api.cloud.ENDPOINT}/api/v2/product-items/{serial_number}/counters/battery/reset",
                "",
                HEADERS(self._api.access_token),
            )
            self._fetch(True)

    def reset_blade_counter(self, serial_number: str) -> None:
        """Resets blade counter.

        Args:
            serial_number (str): Serial number of the device
            data (str): Data to be sent, formatted as a valid JSON object.

        Raises:
            OfflineError: Raised if the device isn't online.
        """
        mower = self.get_mower(serial_number)
        if mower["online"]:
            _LOGGER.debug("Resetting blade counter for %s", mower["name"])
            self._api.check_token()
            POST(
                f"https://{self._api.cloud.ENDPOINT}/api/v2/product-items/{serial_number}/counters/blade/reset",
                "",
                HEADERS(self._api.access_token),
            )
            self._fetch(True)

    def get_cutting_height(self, serial_number: str) -> int:
        """Get the current cutting height of the device.

        Args:
            serial_number (str): Serial number of the device

        Returns:
            int: Cutting height in mm

        Raises:
            NoCuttingHeightError: Raised if the device does not support cutting height.
        """
        mower = self.get_mower(serial_number)
        try:
            return int(mower["last_status"]["payload"]["cfg"]["modules"]["EA"]["h"])
        except KeyError:
            raise NoCuttingHeightError("This device does not support cutting height")

    def set_cutting_height(self, serial_number: str, height: int) -> None:
        """Set the cutting height of the device.

        Args:
            serial_number (str): Serial number of the device
            height (int): Cutting height in mm

        Raises:
            NoCuttingHeightError: Raised if the device does not support cutting height.
            OfflineError: Raised if the device is offline.
        """
        mower = self.get_mower(serial_number)
        if mower["online"]:
            device = DeviceHandler(self._api, mower, self._tz)
            if device.capabilities.check(DeviceCapability.CUTTING_HEIGHT):
                self.mqtt.publish(
                    serial_number if mower["protocol"] == 0 else mower["uuid"],
                    mower["mqtt_topics"]["command_in"],
                    {"cmd": 0, "modules": {"EA": {"h": height}}},
                    mower["protocol"],
                )
            else:
                raise NoCuttingHeightError(
                    "This device does not support cutting height"
                )
        else:
            raise OfflineError("The device is currently offline, no action was sent.")

    def set_acs(self, serial_number: str, state: bool) -> None:
        """Enable or disable the ACS module.

        Args:
            serial_number (str): Serial number of the device
            state (bool): True is enabling ACS, False is disabling ACS.

        Raises:
            NoACSModuleError: Raised if the device does not support ACS.
            OfflineError: Raised if the device is offline.
        """
        mower = self.get_mower(serial_number)
        if mower["online"]:
            device = DeviceHandler(self._api, mower, self._tz)
            if device.capabilities.check(DeviceCapability.ACS):
                self.mqtt.publish(
                    serial_number if mower["protocol"] == 0 else mower["uuid"],
                    mower["mqtt_topics"]["command_in"],
                    {"cmd": 0, "modules": {"US": {"enabled": 1 if state else 0}}},
                    mower["protocol"],
                )
            else:
                raise NoACSModuleError(
                    "This device does not have an ACS module installed."
                )
        else:
            raise OfflineError("The device is currently offline, no action was sent.")



"""MQTT information class."""

from __future__ import annotations

import asyncio
import json
import logging
import random
import ssl
import time
import urllib.parse
from datetime import datetime
from logging import Logger
from typing import Any
from uuid import uuid4

import paho.mqtt.client as mqtt
from paho.mqtt.client import connack_string

from ..events import EventHandler, LandroidEvent
from ..exceptions import NoConnectionError
from .landroid_class import LDict

QOS_FLAG = 1


class MQTTMsgType(LDict):
    """Define specific message type data."""

    def __init__(self) -> dict:
        super().__init__()

        self["in"] = 0
        self["out"] = 0


class MQTTMessageItem(LDict):
    """Defines a MQTT message for Landroid Cloud."""

    def __init__(
        self, device: str, data: str = "{}", qos: int = 0, retain: bool = False
    ) -> dict:
        super().__init__()

        self["device"] = device
        self["data"] = data
        self["qos"] = qos
        self["retain"] = retain


class MQTTMessages(LDict):
    """Messages class."""

    def __init__(self) -> dict:
        super().__init__()

        self["raw"] = MQTTMsgType()
        self["filtered"] = MQTTMsgType()


class MQTTTopics(LDict):
    """Topics class."""

    def __init__(
        self, topic_in: str | None = None, topic_out: str | None = None
    ) -> dict:
        super().__init__()

        self["in"] = topic_in
        self["out"] = topic_out


class Command:
    """Landroid Cloud commands."""

    FORCE_REFRESH = 0
    START = 1
    PAUSE = 2
    HOME = 3
    ZONETRAINING = 4
    LOCK = 5
    UNLOCK = 6
    RESTART = 7
    PAUSE_OVER_WIRE = 8
    SAFEHOME = 9


class MQTT(LDict):
    """Full MQTT handler class."""

    def __init__(
        self,
        api: Any,
        brandprefix: str,
        endpoint: str,
        user_id: int,
        logger: Logger,
        callback: Any,
    ) -> dict:
        """Initialize AWSIoT MQTT handler."""

        super().__init__()
        # self.client = None
        self._events = EventHandler()
        self._on_update = callback
        self._endpoint = endpoint
        self._log = logger.getChild("MQTT")
        self._reconnected: bool = False
        self._topic: list = []
        self._api = api
        self._await_publish: bool = False
        self._await_timestamp: time = None
        self._uuid = uuid4()
        self._is_connected: bool = False
        self._brandprefix = brandprefix
        self._user_id = user_id

        self.client = mqtt.Client(
            client_id=f"{self._brandprefix}/USER/{self._user_id}/homeassistant/{self._uuid}",
            clean_session=False,
            userdata=None,
            reconnect_on_failure=True,
        )

        accesstokenparts = (
            api.access_token.replace("_", "/").replace("-", "+").split(".")
        )
        self.client.username_pw_set(
            username=f"bot?jwt={urllib.parse.quote(accesstokenparts[0])}.{urllib.parse.quote(accesstokenparts[1])}&x-amz-customauthorizer-name=''&x-amz-customauthorizer-signature={urllib.parse.quote(accesstokenparts[2])}",  # pylint: disable= line-too-long
            password=None,
        )

        ssl_context = ssl.create_default_context()
        ssl_context.set_alpn_protocols(["mqtt"])
        self.client.tls_set_context(context=ssl_context)

        self.client.reconnect_delay_set(min_delay=30, max_delay=300)

        self.client.on_connect = self._on_connect
        self.client.on_message = self._forward_on_message
        self.client.on_disconnect = self._on_disconnect

    @property
    def connected(self) -> bool:
        """Returns the MQTT connection state."""
        return self._is_connected
        # return self.client.is_connected()

    def _forward_on_message(
        self,
        client: mqtt.Client | None,  # pylint: disable=unused-argument
        userdata: Any | None,  # pylint: disable=unused-argument
        message: Any | None,
        properties: Any | None = None,  # pylint: disable=unused-argument
    ) -> None:
        """MQTT callback method definition."""
        msg = message.payload.decode("utf-8")
        self._log.debug("Received MQTT message:\n%s", msg)
        self._await_publish = False
        self._on_update(msg)

    def subscribe(self, topic: str, append: bool = True) -> None:
        """Subscribe to MQTT updates."""
        if append and topic not in self._topic:
            self._topic.append(topic)
        self.client.subscribe(topic=topic, qos=QOS_FLAG)

    def connect(self) -> None:
        """Connect to the MQTT service."""
        try:
            self.client.connect(
                self._endpoint,
                443,
            )
            self.client.loop_start()
            while not self.connected:
                try:
                    loop = asyncio.get_running_loop()
                except (
                    RuntimeError
                ):  # 'RuntimeError: There is no current event loop...'
                    loop = None

                if loop and loop.is_running():
                    loop.create_task(asyncio.sleep(0.5))
                else:
                    asyncio.run(asyncio.sleep(0.5))
        except NoConnectionError as exc:
            raise NoConnectionError() from exc

    def _on_connect(
        self,
        client: mqtt.Client | None,  # pylint: disable=unused-argument
        userdata: Any | None,  # pylint: disable=unused-argument
        flags: Any | None,  # pylint: disable=unused-argument
        rc: int | None,
        properties: Any | None = None,  # pylint: disable=unused-argument,invalid-name
    ) -> None:
        """MQTT callback method."""
        logger = self._log.getChild("Conn_State")
        logger.debug(connack_string(rc))
        if rc == 0:
            self._is_connected = True
            self._reconnected = False
            logger.debug("MQTT connected")
            self._events.call(
                LandroidEvent.MQTT_CONNECTION, state=self.client.is_connected()
            )
            for topic in self._topic:
                logger.debug("Subscribing '%s'", topic)
                self.subscribe(topic, False)
            self._await_publish = False
        else:
            self._is_connected = False
            logger.debug("MQTT connection failed")
            self._events.call(
                LandroidEvent.MQTT_CONNECTION, state=self.client.is_connected()
            )

    def _on_disconnect(
        self,
        client: mqtt.Client | None,  # pylint: disable=unused-argument
        userdata: Any | None,  # pylint: disable=unused-argument
        rc: int | None,
        properties: Any | None = None,  # pylint: disable=unused-argument,invalid-name
    ) -> None:
        """MQTT callback method."""
        logger = self._log.getChild("Conn_State")
        self._is_connected = False
        if rc > 0:
            if rc == 7:
                # if not self._reconnected:
                self._reconnected = True
                logger.debug("Reconnecting MQTT")
                self._api.check_token()
                accesstokenparts = (
                    self._api.access_token.replace("_", "/")
                    .replace("-", "+")
                    .split(".")
                )
                self.client.username_pw_set(
                    username=f"bot?jwt={urllib.parse.quote(accesstokenparts[0])}.{urllib.parse.quote(accesstokenparts[1])}&x-amz-customauthorizer-name=''&x-amz-customauthorizer-signature={urllib.parse.quote(accesstokenparts[2])}",  # pylint: disable= line-too-long
                    password=None,
                )
                # else:
                #     self.disconnect()
                #     raise NoConnectionError("Error connecting to AwSIoT MQTT")
            else:
                logger.debug(
                    "Unexpected MQTT disconnect (%s: %s) - retrying",
                    rc,
                    connack_string(rc),
                )
                try:
                    self.client.reconnect()
                except:  # pylint: disable=bare-except
                    pass

    def update_token(self) -> None:
        """Update the token."""
        self._log.debug("Updating token")
        accesstokenparts = (
            self._api.access_token.replace("_", "/").replace("-", "+").split(".")
        )
        self.client.username_pw_set(
            username=f"bot?jwt={urllib.parse.quote(accesstokenparts[0])}.{urllib.parse.quote(accesstokenparts[1])}&x-amz-customauthorizer-name=''&x-amz-customauthorizer-signature={urllib.parse.quote(accesstokenparts[2])}",  # pylint: disable= line-too-long
            password=None,
        )
        if self.connected:
            self.client.reconnect()
        else:
            self.connect()

        self._log.debug("Token updated")

    def disconnect(
        self, reasoncode=None, properties=None  # pylint: disable=unused-argument
    ):
        """Disconnect from AWSIoT MQTT server."""
        logger = self._log.getChild("MQTT_Disconnect")
        for topic in self._topic:
            logger.debug("Unsubscribing '%s'", topic)
            self.client.unsubscribe(topic)
        self._topic = []
        self.client.loop_stop()
        self.client.disconnect()
        logger.debug("MQTT disconnected")

    def ping(self, serial_number: str, topic: str, protocol: int = 0) -> None:
        """Ping (update) the mower."""
        cmd = {"cmd": Command.FORCE_REFRESH}
        try:
            self._log.debug("Sending '%s' on topic '%s'", cmd, topic)
            self.publish(serial_number, topic, cmd, protocol)
        except NoConnectionError:
            pass

    def command(
        self, serial_number: str, topic: str, action: Command, protocol: int = 0
    ) -> None:
        """Send a specific command to the mower."""
        cmd = self.format_message(serial_number, {"cmd": action}, protocol)
        self._log.debug("Sending '%s' on topic '%s'", cmd, topic)
        self.client.publish(topic, cmd, QOS_FLAG)

    def publish(
        self, serial_number: str, topic: str, message: dict, protocol: int = 0
    ) -> None:
        """Publish message to the mower."""
        if not self.connected:
            self.update_token()
            # raise NoConnectionError("No connection to AwSIoT MQTT")

        while self._await_publish:
            if self._await_timestamp + 30 >= time.time():
                self._await_publish = False
                break
            asyncio.run(asyncio.sleep(1))

        self._await_publish = True
        self._await_timestamp = time.time()
        self._log.debug("Publishing message '%s'", message)
        self.client.publish(
            topic, self.format_message(serial_number, message, protocol), QOS_FLAG
        )

    def format_message(self, serial_number: str, message: dict, protocol: int) -> str:
        """
        Format a message.
        Message is expected to be a dict like this: {"cmd": 1}
        """
        now = datetime.now()
        msg = {}
        if protocol == 0:
            msg = {
                "id": random.randint(1024, 65535),
                "sn": serial_number,
                "tm": now.strftime("%H:%M:%S"),
                "dt": now.strftime("%d/%m/%Y"),
            }
        elif protocol == 1:
            msg = {
                "id": random.randint(1024, 65535),
                "uuid": serial_number,
                "tm": now.strftime("%Y-%m-%dT%H:%M:%SZ"),
            }

        msg.update(message)
        self._log.debug("Formatting message '%s' to '%s'", message, msg)

        return json.dumps(msg)



"""Class for handling device info and states."""

from __future__ import annotations

import json
import logging
from datetime import datetime, timedelta
from typing import Any

from pyworxcloud.day_map import DAY_MAP

from ..const import UNWANTED_ATTRIBS
from ..exceptions import APIException, InvalidDataDecodeException
from ..helpers import convert_to_time
from .battery import Battery
from .blades import Blades
from .capability import Capability, DeviceCapability
from .firmware import Firmware
from .landroid_class import LDict
from .lawn import Lawn
from .location import Location
from .orientation import Orientation
from .rainsensor import Rainsensor
from .schedules import TYPE_TO_STRING, Schedule, ScheduleType, Weekdays
from .state import States, StateType
from .statistics import Statistic
from .warranty import Warranty
from .zone import Zone

LOGGER = logging.getLogger(__name__)


class DeviceHandler(LDict):
    """DeviceHandler for Landroid Cloud devices."""

    __is_decoded: bool = True
    __raw_data: str = None
    __json_data: str = None

    def __init__(
        self,
        api: Any = None,
        mower: Any = None,
        tz: str | None = None,
        decode: bool = True,
    ) -> dict:
        """Initialize the object."""
        super().__init__()

        self._api = api
        self.mower = mower
        self._tz = tz
        self._decode = decode

        self.battery = Battery()
        self.blades = Blades()
        self.error = States(StateType.ERROR)
        self.orientation = Orientation([0, 0, 0])
        self.capabilities = Capability()
        self.rainsensor = Rainsensor()
        self.status = States()
        self.zone = Zone()
        self.warranty = Warranty()
        self.firmware = Firmware()
        self.schedules = Schedule()
        self.statistics = Statistic([])
        self.in_topic = None
        self.out_topic = None

        if not isinstance(mower, type(None)) and not isinstance(api, type(None)):
            self.__mapinfo(api, mower)

    @property
    def raw_data(self) -> str:
        """Returns current raw dataset."""
        return self.__raw_data

    @property
    def json_data(self) -> str:
        """Returns current dataset as JSON."""
        return self.__json_data

    @raw_data.setter
    def raw_data(self, value: str) -> None:
        """Set new MQTT data."""
        self.__is_decoded = False
        self.__raw_data = value
        try:
            self.__json_data = json.loads(value)
        except:  # pylint: disable=bare-except
            pass  # Just continue if we couldn't decode the data

        self.decode_data()

    @property
    def is_decoded(self) -> bool:
        """Returns true if latest dataset was decoded and handled."""
        return self.__is_decoded

    @is_decoded.setter
    def is_decoded(self, value: bool) -> None:
        """Set decoded flag when dataset was decoded and handled."""
        self.__is_decoded = value

    def __mapinfo(self, api: Any, data: Any) -> None:
        """Map information from API."""

        if isinstance(data, type(None)) or isinstance(api, type(None)):
            raise APIException(
                "Either 'data' or 'api' object was missing, no data was mapped!"
            )

        for attr, val in data.items():
            setattr(self, str(attr), val)

        if not "time_zone" in data:
            data["time_zone"] = "UTC"

        self.battery = Battery(data)
        self.blades = Blades(data)
        self.error = States(StateType.ERROR)
        self.orientation = Orientation([0, 0, 0])
        self.capabilities = Capability(data)
        self.rainsensor = Rainsensor()
        self.status = States()
        self.zone = Zone(data)
        self.warranty = Warranty(data)
        self.firmware = Firmware(data)
        self.schedules = Schedule(data)
        self.statistics = Statistic([])
        self.in_topic = data["mqtt_topics"]["command_in"]
        self.out_topic = data["mqtt_topics"]["command_out"]

        if data in ["lawn_perimeter", "lawn_size"]:
            self.lawn = Lawn(data["lawn_perimeter"], data["lawn_size"])

        self.name = (
            data["name"] if not isinstance(data["name"], type(None)) else "No Name"
        )
        self.model = str.format(
            "{} ({})", data["model"]["friendly_name"], data["model"]["code"]
        )

        self.mac_address = None
        self.protocol = 0
        self.time_zone = None

        for attr in UNWANTED_ATTRIBS:
            if hasattr(self, attr):
                delattr(self, attr)

        if self._decode:
            self.decode_data()
            self.is_decoded = True

    def decode_data(self) -> None:
        """Decode incoming JSON data."""
        invalid_data = False
        self.is_decoded = False

        logger = LOGGER.getChild("decode_data")
        logger.debug("Data decoding for %s started", self.name)

        if self.json_data:
            logger.debug("Found JSON decoded data: %s", self.json_data)
            data = self.json_data
        elif self.raw_data:
            logger.debug("Found raw data: %s", self.raw_data)
            data = self.raw_data
        elif (
            not isinstance(self.last_status, type(None))
            and "payload" in self.last_status
        ):
            data = self.last_status["payload"]
        else:
            self.is_decoded = True
            logger.debug("No valid data was found, skipping update for %s", self.name)
            return

        if isinstance(self.capabilities, list):
            setattr(self, "api_capabilities", getattr(self, "capabilities"))
            self.capabilities = Capability(data)

        mower = self.mower
        self.protocol = mower["protocol"]

        if "dat" in data:
            mower["last_status"]["payload"]["dat"] = data["dat"]
            if "uuid" in data["dat"]:
                self.uuid = data["dat"]["uuid"]

            if isinstance(self.mac_address, type(None)):
                self.mac_address = (
                    data["dat"]["mac"] if "mac" in data["dat"] else "__UUID__"
                )

            try:
                # Get wifi signal strength
                if "rsi" in data["dat"]:
                    self.rssi = data["dat"]["rsi"]

                # Get status code
                if "ls" in data["dat"]:
                    self.status.update(data["dat"]["ls"])

                # Get error code
                if "le" in data["dat"]:
                    self.error.update(data["dat"]["le"])

                # Get zone index
                self.zone.index = data["dat"]["lz"] if "lz" in data["dat"] else 0

                # Get device lock state
                self.locked = bool(data["dat"]["lk"]) if "lk" in data["dat"] else None
                mower["locked"] = self.locked

                # Get battery info if available
                if "bt" in data["dat"]:
                    if len(self.battery) == 0:
                        self.battery = Battery(data["dat"]["bt"])
                    else:
                        self.battery.set_data(data["dat"]["bt"])
                # Get device statistics if available
                if "st" in data["dat"]:
                    self.statistics = Statistic(data["dat"]["st"])

                    if len(self.blades) != 0:
                        self.blades.set_data(data["dat"]["st"])

                # Get orientation if available.
                if "dmp" in data["dat"]:
                    self.orientation = Orientation(data["dat"]["dmp"])

                # Check for extra module availability
                if "modules" in data["dat"]:
                    if "4G" in data["dat"]["modules"]:
                        if "gps" in data["dat"]["modules"]["4G"]:
                            self.gps = Location(
                                data["dat"]["modules"]["4G"]["gps"]["coo"][0],
                                data["dat"]["modules"]["4G"]["gps"]["coo"][1],
                            )

                # Get remaining rain delay if available
                if "rain" in data["dat"]:
                    self.rainsensor.triggered = bool(
                        str(data["dat"]["rain"]["s"]) == "1"
                    )
                    self.rainsensor.remaining = int(data["dat"]["rain"]["cnt"])

            except TypeError:  # pylint: disable=bare-except
                invalid_data = True

        if "cfg" in data:
            mower["last_status"]["payload"]["cfg"] = data["cfg"]
            # try:
            if "dt" in data["cfg"]:
                dt_split = data["cfg"]["dt"].split("/")
                date = (
                    f"{dt_split[2]}-{dt_split[1]}-{dt_split[0]}"
                    + " "
                    + data["cfg"]["tm"]
                )
            elif "tm" in data["dat"]:
                date = datetime.fromisoformat(data["dat"]["tm"])
            else:
                date = datetime.now()

            self.updated = date
            self.rainsensor.delay = int(data["cfg"]["rd"])

            # Fetch wheel torque
            if "tq" in data["cfg"]:
                self.capabilities.add(DeviceCapability.TORQUE)
                self.torque = data["cfg"]["tq"]

            # Fetch zone information
            if "mz" in data["cfg"] and "mzv" in data["cfg"]:
                self.zone.starting_point = data["cfg"]["mz"]
                self.zone.indicies = data["cfg"]["mzv"]

                # Map current zone to zone index
                self.zone.current = self.zone.indicies[self.zone.index]

            # Fetch main schedule
            if "sc" in data["cfg"]:
                if "ots" in data["cfg"]["sc"]:
                    self.capabilities.add(DeviceCapability.ONE_TIME_SCHEDULE)
                    self.capabilities.add(DeviceCapability.EDGE_CUT)
                if "m" in data["cfg"]["sc"] or "enabled" in data["cfg"]["sc"]:
                    self.capabilities.add(DeviceCapability.PARTY_MODE)
                    self.partymode_enabled = (
                        bool(str(data["cfg"]["sc"]["m"]) == "2")
                        if self.protocol == 0
                        else bool(str(data["cfg"]["sc"]["enabled"]) == "0")
                    )
                    self.schedules["active"] = (
                        bool(str(data["cfg"]["sc"]["m"]) in ["1", "2"])
                        if self.protocol == 0
                        else bool(str(data["cfg"]["sc"]["enabled"]) == "0")
                    )

                self.schedules["time_extension"] = (
                    data["cfg"]["sc"]["p"] if self.protocol == 0 else "0"
                )

                sch_type = ScheduleType.PRIMARY
                self.schedules.update({TYPE_TO_STRING[sch_type]: Weekdays()})

                try:
                    for day in range(
                        0,
                        (
                            len(data["cfg"]["sc"]["d"])
                            if self.protocol == 0
                            else len(data["cfg"]["sc"]["slots"])
                        ),
                    ):
                        dayOfWeek = (  # pylint: disable=invalid-name
                            day
                            if self.protocol == 0
                            else data["cfg"]["sc"]["slots"][day]["d"]
                        )
                        self.schedules[TYPE_TO_STRING[sch_type]][DAY_MAP[dayOfWeek]][
                            "start"
                        ] = (
                            data["cfg"]["sc"]["d"][day][0]
                            if self.protocol == 0
                            else (
                                datetime.strptime("00:00", "%H:%M")
                                + timedelta(
                                    minutes=data["cfg"]["sc"]["slots"][day]["s"]
                                )
                            ).strftime("%H:%M")
                        )
                        self.schedules[TYPE_TO_STRING[sch_type]][DAY_MAP[dayOfWeek]][
                            "duration"
                        ] = (
                            data["cfg"]["sc"]["d"][day][1]
                            if self.protocol == 0
                            else data["cfg"]["sc"]["slots"][day]["t"]
                        )
                        self.schedules[TYPE_TO_STRING[sch_type]][DAY_MAP[dayOfWeek]][
                            "boundary"
                        ] = (
                            bool(data["cfg"]["sc"]["d"][day][2])
                            if self.protocol == 0
                            else (
                                bool(data["cfg"]["sc"]["slots"][day]["cfg"]["cut"]["b"])
                                if "b" in data["cfg"]["sc"]["slots"][day]["cfg"]["cut"]
                                else None
                            )
                        )

                        time_start = datetime.strptime(
                            self.schedules[TYPE_TO_STRING[sch_type]][
                                DAY_MAP[dayOfWeek]
                            ]["start"],
                            "%H:%M",
                        )

                        if isinstance(
                            self.schedules[TYPE_TO_STRING[sch_type]][
                                DAY_MAP[dayOfWeek]
                            ]["duration"],
                            type(None),
                        ):
                            self.schedules[TYPE_TO_STRING[sch_type]][
                                DAY_MAP[dayOfWeek]
                            ]["duration"] = "0"

                        duration = int(
                            self.schedules[TYPE_TO_STRING[sch_type]][
                                DAY_MAP[dayOfWeek]
                            ]["duration"]
                        )

                        duration = duration * (
                            1 + (int(self.schedules["time_extension"]) / 100)
                        )
                        end_time = time_start + timedelta(minutes=duration)

                        self.schedules[TYPE_TO_STRING[sch_type]][DAY_MAP[dayOfWeek]][
                            "end"
                        ] = end_time.time().strftime("%H:%M")
                except KeyError:
                    pass

                # Fetch secondary schedule
                try:
                    if "dd" in data["cfg"]["sc"]:
                        sch_type = ScheduleType.SECONDARY
                        self.schedules.update({TYPE_TO_STRING[sch_type]: Weekdays()})

                        for day in range(0, len(data["cfg"]["sc"]["dd"])):
                            self.schedules[TYPE_TO_STRING[sch_type]][DAY_MAP[day]][
                                "start"
                            ] = data["cfg"]["sc"]["dd"][day][0]
                            self.schedules[TYPE_TO_STRING[sch_type]][DAY_MAP[day]][
                                "duration"
                            ] = data["cfg"]["sc"]["dd"][day][1]
                            self.schedules[TYPE_TO_STRING[sch_type]][DAY_MAP[day]][
                                "boundary"
                            ] = bool(data["cfg"]["sc"]["dd"][day][2])

                            time_start = datetime.strptime(
                                data["cfg"]["sc"]["dd"][day][0],
                                "%H:%M",
                            )

                            if isinstance(
                                self.schedules[TYPE_TO_STRING[sch_type]][DAY_MAP[day]][
                                    "duration"
                                ],
                                type(None),
                            ):
                                self.schedules[TYPE_TO_STRING[sch_type]][DAY_MAP[day]][
                                    "duration"
                                ] = "0"

                            duration = int(
                                self.schedules[TYPE_TO_STRING[sch_type]][DAY_MAP[day]][
                                    "duration"
                                ]
                            )

                            duration = duration * (
                                1 + (int(self.schedules["time_extension"]) / 100)
                            )
                            end_time = time_start + timedelta(minutes=duration)

                            self.schedules[TYPE_TO_STRING[sch_type]][DAY_MAP[day]][
                                "end"
                            ] = end_time.time().strftime("%H:%M")
                except KeyError:
                    pass

                # Check for addon modules
                if "modules" in data["cfg"]:
                    if "DF" in data["cfg"]["modules"]:
                        self.capabilities.add(DeviceCapability.OFF_LIMITS)
                        self.offlimit = bool(
                            str(data["cfg"]["modules"]["DF"]["cut"]) == "1"
                        )
                        self.offlimit_shortcut = bool(
                            str(data["cfg"]["modules"]["DF"]["fh"]) == "1"
                        )

                    if "US" in data["cfg"]["modules"]:
                        self.capabilities.add(DeviceCapability.ACS)
                        self.acs_enabled = bool(
                            str(data["cfg"]["modules"]["US"]["enabled"]) == "1"
                        )

            self.schedules.update_progress_and_next(
                tz=(
                    self._tz if not isinstance(self._tz, type(None)) else self.time_zone
                )
            )
            # except TypeError:
            #     invalid_data = True
            # except KeyError:
            #     invalid_data = True

        convert_to_time(self.name, self, self._tz, callback=self.update_attribute)

        mower["last_status"]["timestamp"] = self.updated

        self.is_decoded = True
        logger.debug("Data for %s was decoded", self.name)
        logger.debug("Device object:\n%s", vars(self))

        if invalid_data:
            raise InvalidDataDecodeException()

    def update_attribute(self, device: str, attr: str, key: str, value: Any) -> None:
        """Used as callback to update value."""
        chattr = self
        if not isinstance(attr, type(None)):
            for level in attr.split(";;"):
                if hasattr(chattr, level):
                    chattr = getattr(chattr, level)
                else:
                    chattr = chattr[level]

        if hasattr(chattr, key):
            setattr(chattr, key, value)
        elif isinstance(chattr, dict):
            chattr.update({key: value})



"""Defines schedule classes."""

from __future__ import annotations

from datetime import datetime, timedelta
from enum import IntEnum
from typing import Any
from zoneinfo import ZoneInfo

from ..day_map import DAY_MAP
from .landroid_class import LDict


class ScheduleType(IntEnum):
    """Schedule types."""

    PRIMARY = 0
    SECONDARY = 1


TYPE_TO_STRING = {ScheduleType.PRIMARY: "primary", ScheduleType.SECONDARY: "secondary"}


class WeekdaySettings(LDict):
    """Class representing a weekday setting."""

    def __init__(
        self,
        start: str = "00:00",
        end: str = "00:00",
        duration: int = 0,
        boundary: bool = False,
    ) -> None:
        """Initialize the settings."""
        super().__init__()
        self["start"] = start
        self["end"] = end
        self["duration"] = duration
        self["boundary"] = boundary


class Weekdays(LDict):
    """Represents all weekdays."""

    def __init__(self) -> dict:
        super().__init__()

        for day in [
            "monday",
            "tuesday",
            "wednesday",
            "thursday",
            "friday",
            "saturday",
            "sunday",
        ]:
            # Initialize each weekday with default settings
            self.update({day.lower(): WeekdaySettings()})


class ScheduleInfo:
    """Used for calculate the current schedule progress and show next schedule start."""

    def __init__(self, schedule: Schedule, tz: str | None = None) -> None:
        """Initialize the ScheduleInfo object and set values."""
        self.__schedule = schedule
        now = datetime.now()
        timezone = ZoneInfo(tz) if not isinstance(tz, type(None)) else ZoneInfo("UTC")
        self._tz = tz
        self.__now = now.astimezone(timezone)
        self.__today = self.__now.strftime("%d/%m/%Y")

    def _get_schedules(
        self, date: datetime, next: bool = False, add_offset: bool = False
    ) -> WeekdaySettings | None:
        """Get primary and secondary schedule for today or tomorrow."""
        day = DAY_MAP[int(date.strftime("%w"))]

        primary = self.__schedule[TYPE_TO_STRING[ScheduleType.PRIMARY]][day]
        if primary["duration"] == 0:
            return None, None, date

        secondary = (
            self.__schedule[TYPE_TO_STRING[ScheduleType.SECONDARY]][day]
            if TYPE_TO_STRING[ScheduleType.SECONDARY] in self.__schedule
            else None
        )

        if (not isinstance(secondary, type(None))) and secondary["duration"] == 0:
            secondary = None

        return primary, secondary, date

    def calculate_progress(self) -> int:
        """Calculate and return current progress in percent."""
        from ..helpers.time_format import string_to_time

        primary, secondary, date = self._get_schedules(self.__now)

        if isinstance(primary, type(None)) and isinstance(secondary, type(None)):
            return 100

        start = string_to_time(f"{self.__today} {primary['start']}:00", self._tz)
        end = string_to_time(f"{self.__today} {primary['end']}:00", self._tz)

        total_run = primary["duration"]
        has_run = 0

        if self.__now >= start and self.__now < end:
            has_run = (self.__now - start).total_seconds() / 60
        elif self.__now < start:
            has_run = 0
        else:
            has_run = primary["duration"]

        if (not isinstance(secondary, type(None))) and secondary["duration"] > 0:
            start = string_to_time(f"{self.__today} {secondary['start']}:00", self._tz)
            end = string_to_time(f"{self.__today} {secondary['end']}:00", self._tz)
            total_run += secondary["duration"]
            if self.__now >= start and self.__now < end:
                has_run += (self.__now - start).total_seconds() / 60
            elif self.__now < start:
                has_run += 0
            else:
                has_run += primary["duration"]

        pct = (has_run / total_run) * 100
        return int(round(pct))

    def next_schedule(self) -> str:
        """Find next schedule starting point."""
        from ..helpers.time_format import string_to_time

        primary, secondary, date = self._get_schedules(self.__now)
        next = None
        cnt = 0
        while isinstance(primary, type(None)):
            if cnt == 7:
                # No schedule active for any weekday, return None
                return None

            primary, secondary, date = self._get_schedules(date + timedelta(days=1))
            cnt += 1

        start = string_to_time(
            f"{date.strftime('%d/%m/%Y')} {primary['start']}:00", self._tz
        )

        if self.__now < start:
            next = start
        elif (
            (not isinstance(secondary, type(None)))
            and start
            < string_to_time(
                f"{date.strftime('%d/%m/%Y')} {secondary['start']}:00", self._tz
            )
            and secondary["duration"] > 0
        ):
            next = string_to_time(
                f"{date.strftime('%d/%m/%Y')} {secondary['start']}:00", self._tz
            )

        if isinstance(next, type(None)):
            primary, secondary, date = self._get_schedules(date + timedelta(days=1))
            while isinstance(primary, type(None)):
                primary, secondary, date = self._get_schedules(date + timedelta(days=1))
            next = string_to_time(
                f"{date.strftime('%d/%m/%Y')} {primary['start']}:00", self._tz
            )

        return next.strftime("%Y-%m-%d %H:%M:%S")


class Schedule(LDict):
    """Represents a schedule."""

    def __init__(self, data: Any | None = None) -> None:
        """Initialize a schedule."""
        if isinstance(data, type(None)):
            return

        super().__init__()

        self["daily_progress"] = None
        self["next_schedule_start"] = None
        self["time_extension"] = 0
        self["active"] = True
        self["auto_schedule"] = {
            "settings": (
                data["auto_schedule_settings"]
                if "auto_schedule_settings" in data
                else {}
            ),
            "enabled": data["auto_schedule"] if "auto_schedule" in data else False,
        }

    def update_progress_and_next(self, tz: str | None = None) -> None:
        """Update progress and next scheduled start properties."""

        info = ScheduleInfo(self, tz)
        self["daily_progress"] = info.calculate_progress()
        self["next_schedule_start"] = info.next_schedule()



"""Zone representation."""

from __future__ import annotations

from typing import Any

from .landroid_class import LDict


class Zone(LDict):
    """Class for handling zone data."""

    def __init__(self, data: Any | None = None) -> dict:
        """Initialize zone object."""
        if isinstance(data, type(None)):
            return

        super().__init__()

        self["current"] = 0
        self["index"] = 0
        self["indicies"] = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        self["starting_point"] = [0, 0, 0, 0]

        try:
            if not "last_status" in data:
                return

            if not "payload" in data["last_status"]:
                return

            if (
                not "dat" in data["last_status"]["payload"]
                or not "cfg" in data["last_status"]["payload"]
            ):
                return

            self["index"] = (
                data["last_status"]["payload"]["dat"]["lz"]
                if "lz" in data["last_status"]["payload"]["dat"]
                else 0
            )
            self["indicies"] = (
                data["last_status"]["payload"]["cfg"]["mzv"]
                if "mzv" in data["last_status"]["payload"]["cfg"]
                else [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
            )
            self["starting_point"] = (
                data["last_status"]["payload"]["cfg"]["mz"]
                if "mz" in data["last_status"]["payload"]["cfg"]
                else [0, 0, 0, 0]
            )
        except TypeError:  # pylint: disable=bare-except
            pass

        self["current"] = self["indicies"][self["index"]]

    @property
    def current(self) -> int:
        """Get current zone."""
        return self["current"]

    @current.setter
    def current(self, value: int) -> None:
        """Set current zone property."""
        self["current"] = value

    @property
    def index(self) -> int:
        """Get current index."""
        return self["index"]

    @index.setter
    def index(self, value: int) -> None:
        """Set current index property."""
        self["index"] = value

    @property
    def indicies(self) -> int:
        """Get indicies."""
        return self["indicies"]

    @indicies.setter
    def indicies(self, value: list) -> None:
        """Set indicies property."""
        self["indicies"] = value

    @property
    def starting_point(self) -> int:
        """Get starting points."""
        return self["starting_point"]

    @starting_point.setter
    def starting_point(self, value: int) -> None:
        """Set starting points."""
        self["starting_point"] = value
